from pathlib import Path
from typing import Any, Dict, MutableMapping

from .json_parser import json_parser
from .parser import Parser
from .toml_parser import toml_parser

parser_mapping: Dict[str, Parser] = {
    ".json": json_parser,
    ".toml": toml_parser,
}


def auto_parser(filepath: Path) -> MutableMapping[str, Any]:
    """Resolve a parse using the file's extension & use it to construct a dict."""
    resolved_parser = parser_mapping[filepath.suffix]
    return resolved_parser(filepath)



import sys
from abc import abstractmethod
from pathlib import Path
from typing import Any, MutableMapping

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


class Parser(Protocol):
    """Protocol typing for a parsing function."""

    @abstractmethod
    def __call__(self, filepath: Path) -> MutableMapping[str, Any]:
        """Open a file, parse the contents, construct a dict & return it."""



import json
from pathlib import Path
from typing import Any, MutableMapping


def json_parser(filepath: Path) -> MutableMapping[str, Any]:
    """Open, read and construct dict from JSON file."""
    with filepath.open("r") as f:
        return json.load(f)



from os import PathLike
from pathlib import Path
from typing import Any, MutableMapping, Optional, Union

from .exceptions import DoubleCreation
from .parser import Parser, auto_parser


class ClassyConfig:
    """Define your config filepath."""

    instance: Optional["ClassyConfig"] = None
    _raw_config: Optional[MutableMapping[str, Any]] = None

    def __new__(cls, *args, **kwargs) -> "ClassyConfig":
        """
        Singleton implementation.

        if the object has already been created a DoubleCreation error is raised.
        """
        if cls.instance is not None:
            raise DoubleCreation("Can only create one ClassyConfig")

        cls.instance = super(ClassyConfig, cls).__new__(cls)
        return cls.instance

    def __init__(self, filepath: Union[str, PathLike], parser: Parser = auto_parser) -> None:
        self.config_file = Path(filepath)
        self.parser = parser

        if not self.config_file.exists():
            raise FileNotFoundError()

    @property
    def raw_config(self) -> MutableMapping[str, Any]:
        """
        Gathers the raw config from the file.

        This config file is read once & saved to an instance attribute,
            so future access will not require another file read.
        """
        if self._raw_config is None:
            self._raw_config = self.parser(self.config_file)

        return self._raw_config
